#include "../include/ServerUtils.h"
#include "../include/WordUtils.h"
#include "../include/DataType.h"
#include "../include/FileUtils.h"

#include <string>
#include <vector>
#include <iostream>
#include <sstream>
#include <unistd.h>
#include <netinet/in.h>
#include <thread>
#include <chrono>
#include <cstring>
#include <random>

vector<string> css;
vector<string> images;
unordered_map<string, unordered_map<string, int>> word_frequencies;

void ServerUtils::serve() {
    word_frequencies = WordUtils::load_data("/home/skingging/Documents/Projects/CPP/AI-Tarpit-Reimagined/wordlist/wordlist-Food.txt");
    css = WordUtils::load_css("/home/skingging/Documents/Projects/CPP/AI-Tarpit-Reimagined/content/style.css");
    images = FileUtils::get_image_list("/home/skingging/Documents/Projects/CPP/AI-Tarpit-Reimagined/content/");

    // server_fd is a file descriptor.
    const int server_fd = socket(AF_INET, SOCK_STREAM, 0);
    
    sockaddr_in addr{};
    addr.sin_family = AF_INET;
    addr.sin_port = htons(8888);
    addr.sin_addr.s_addr = INADDR_ANY;
    
    bind(server_fd, reinterpret_cast<sockaddr *>(&addr), sizeof(addr));
    listen(server_fd, 15);
    
    cout << "Server is running on http://localhost:8888 \n";
    
    while (true) {
        int client_fd = accept(server_fd, nullptr, nullptr);
        
        thread(process_request, client_fd).detach();
    }
    
    close(server_fd);
}

void ServerUtils::process_request(const int client_fd) {
    char buffer[1024];
    const int bytes_received = recv(client_fd, buffer, sizeof(buffer) - 1, 0);

    string url;

    if (bytes_received > 0) {
        buffer[bytes_received] = '\0';

        url = WordUtils::extract_url(string(buffer));
    } else {
        cerr << "AAAA \n";
    }

    if (url == "/style.css") {
        // This sends the header, that instructs how the browser should interpret the data.
        send_header(client_fd, data_type::CSS);
        send_chunked_css(client_fd);
    }

    else if (url == "/favicon.png") {
        send_image(client_fd, "/home/skingging/Documents/Projects/CPP/AI-Tarpit-Reimagined/a.png", image_type::PNG);
    }

    else if (WordUtils::contains_image(url)) {
        string p = "/home/skingging/Documents/Projects/CPP/AI-Tarpit-Reimagined/content/images/";
        p += WordUtils::extract_image_name(url);
        cerr << p << endl;
        send_image(client_fd, p, image_type::AVIF);
    }

    else {
        //unsigned int hash = WordUtils::hash_url(url);
        //unsigned long hash2 = WordUtils::djb2Hash(url);
        const unsigned long hash3 = WordUtils::fnv1aHash(url);

        send_header(client_fd, HTML);
        send_chunked_html(client_fd, hash3);
    }

    // Send final zero-length chunk to end the response
    send(client_fd, "0\r\n\r\n", 5, 0);
    
    close(client_fd);
}

void ServerUtils::send_header(const int client_fd, const data_type type) {
    if (type == HTML) {
        send_data(client_fd, HTML_RESPONSE_HEADER);
    }
    
    else if (type == CSS) {
        send_data(client_fd, CSS_RESPONSE_HEADER);
    }
}

void ServerUtils::send_chunked_html(const int client_fd, const size_t hash) {
    const string hashes = to_string(hash);

    unsigned short itr = 0;
    const unsigned short end = hashes.size();

    send_data(client_fd, HTML_BEGINNING);
    send_data(client_fd, HTML_NAV);
    send_data(client_fd, HTML_MAIN_1);

    minstd_rand generator(hash);
    uniform_int_distribution<unsigned short> distribution_1(0, end - 1);
    uniform_int_distribution<unsigned short> distribution_2(0, images.size() - 1);

    const int link = distribution_1(generator);
    const int image = distribution_2(generator);

    while (itr < end) {
        send_data(client_fd, WordUtils::create_tag(word_frequencies, hashes[itr]));

        if (itr == link) {
            send_data(client_fd, WordUtils::create_link(word_frequencies, hash));
        }

        if (itr == link) {
            send_data(client_fd, WordUtils::create_image(images[image]));
        }

        this_thread::sleep_for(chrono::milliseconds(25));

        itr++;
    }

    send_data(client_fd, HTML_MAIN_2);
    send_data(client_fd, HTML_FOOTER);
    send_data(client_fd, HTML_END);
}

void ServerUtils::send_chunked_css(const int client_fd) {
    for (size_t i = 0; i < css.size(); i++) {
        send_data(client_fd, css[i]);

        this_thread::sleep_for(chrono::milliseconds(25));
    }
}

void ServerUtils::send_image(const int client_fd, const string& path, const image_type type) {
    const vector<unsigned char> image = FileUtils::open_image(path);

    std::ostringstream oss;
    oss << "HTTP/1.1 200 OK\r\n";
    if (type == PNG) {
        oss << "Content-Type: image/png\r\n";
    }

    if (type == AVIF) {
        oss << "Content-Type: image/avif\r\n";
    }

    oss << "Content-Length: " << image.size() << "\r\n";
    oss << "\r\n";

    //send(client_fd, oss.str().c_str(), oss.str().size(), 0);
    send_data(client_fd, oss.str());

    send(client_fd, reinterpret_cast<const char*>(image.data()), image.size(), 0);
    //send_all(client_fd,reinterpret_cast<const char*>(image.data()), image.size());
}

size_t ServerUtils::send_all(int sockfd, const char *data, const size_t length) {
    size_t total_sent = 0;
    while (total_sent < length) {
        size_t sent = send(sockfd, data + total_sent, length - total_sent, 0);
        if (sent <= 0) {
            return sent;  // Error or connection closed
        }
        total_sent += sent;
    }
    return total_sent;
}

void ServerUtils::send_data(const int client_fd, const string& data) {
    ostringstream oss;

    oss << hex << data.size() << "\r\n" << data << "\r\n";

    const int result = send(client_fd, oss.str().c_str(), oss.str().size(), 0);

    if (result == -1) {
        return;
    }
}